{% extends "base.html" %}

{% block title %}
	&raquo; Supplier Contracts &raquo; {{contract.name}} &raquo; Batches &raquo; Add
{% endblock %}

{% block inside_head %}
	<style>
		label {
			width: 6em;
			float: left;
		}
	</style>
{% endblock %}

{% block nav %}
	<a href="/e/supplier_contracts">Supplier Contracts</a> &raquo; 
	<a href="/e/supplier_contracts/{{contract.id}}">{{contract.name}}</a> &raquo; 
	<a href="/e/supplier_batches?supplier_contract_id={{contract.id}}">Batches</a> &raquo; 
	Add
{% endblock %}

{% block content %}
	<form action="/e/supplier_contracts/{{contract.id}}/add_batch" method="post">
		<fieldset>
			<legend>Add a batch</legend>	
			<label>Reference</label>
			{{input_text('reference', next_batch_reference, required=True)}}
			<label>Description</label>
			{{ input_text('description', '', placeholder=next_batch_description) }}
			<input type="submit" value="Add">
		</fieldset>
	</form>

	<table>
		<caption>Existing Batches</caption>
		<thead>
			<tr>
				<th>Reference</th>
				<th>Description</th>
			</tr>
		</thead>
		<tbody>
			{% for batch in batches %}
				<tr>
					<td>{{batch.reference}}</td>
					<td>{{batch.description}}</td>
				</tr>
			{% endfor %}
		</tbody>
	</table>

{% endblock %}
