{% extends "base.html" %}

{% block title %}
	&raquo; Parties
{% endblock %}

{% block nav %}
	Parties
{% endblock %}

{% block content %}
	<table>
		<thead>
			<th>Name</th>
			<th>Market Role</th>
			<th>Participant</th>
		</thead>
		<tbody>
			{% for party in parties %}
				<tr>
					<td>
						<a href="/e/parties/{{party.id}}">{{party.name}}</a>
					</td>
					<td>
						<a href="/e/market_roles/{{party.market_role.id}}"
							>{{party.market_role.description}}</a>
					</td>
					<td>
						<a href="/e/participants/{{party.participant.id}}"
							>{{party.participant.code}}</a>
					</td>
				</tr>
			{% endfor %}
		</tbody>
	</table>
{% endblock %}
