/* Etheling */

:root {
  --main-color: #009FD9;
  --fg-color: #333;
	--bg-color: #D9D9D9;
}

body {
	display: grid;
 	grid-template-columns: 1fr 4fr;
 	grid-template-rows: 84px 1fr;
 	font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif;
 	color: var(--fg-color);
	background-color: #EEF4F9;
	font-size: 14px;
	margin: 0;
	line-height: 1.42857143;
}

a {
	color: var(--main-color);
	text-decoration: none;
	transition: background-color 230ms ease, color 230ms ease;
}

a:hover {
	color: #337AB7;
	text-decoration: underline;
}


body > header {
	background-color: var(--main-color);
 	display: grid;
 	grid-template-columns: auto 1fr;
 	grid-template-rows: 1fr;
	align-items: center;
}

body > header > nav {
	color: white;
  position: relative;
  display: inline-block;
	text-align: center;
	padding-left: 1.5em;
}

body > header > nav > span {
	cursor: pointer;
	font-size: large;
}

body > header > nav > nav {
	color: var(--fg-color);
  display: none;
  position: absolute;
	background-color: var(--bg-color);
  min-width: 21em;
  box-shadow: 0px 8px 16px 0px rgba(0,0,0,0.2);
  z-index: 1;
  padding-top: 0.5em;
  padding-bottom: 0.5em;
	left: 0em;
	list-style-type: none;
	padding-left: 0;
	padding-right: 0;
	margin: 0;
	text-align: start;
}

body > header > nav > nav > details {
	cursor: pointer;
	margin: 0;
}

body > header > nav > nav > details > summary {
	list-style: none;
	padding-left: 2em;
	padding-top: 0.5em;
	padding-bottom: 0.5em;
}

body > header > nav > nav > details > ul {
	list-style-type: none;
	padding-left: 0;
	margin: 0;
	color: var(--fg-color);
}

body > header > nav > nav > details > ul > li {
	color: var(--fg-color);
	margin: 0;
	padding-top: 0.5em;
	padding-bottom: 0.5em;
	padding-left: 2em;
	padding-right: 2em;
	background-color: white;
}

body > header > nav > nav > details > ul > li > a:link {
	color: var(--fg-color);
	text-decoration: none;
}

body > header > nav > nav > details > ul > li > a:visited {
	color: var(--fg-color);
	text-decoration: none;
}

body > header > nav > nav > details > ul > li:hover {
	padding-top: 0.5em;
	padding-bottom: 0.5em;
	padding-left: 2em;
	padding-right: 2em;
	background-color: var(--main-color);
	color: white;
}


body > header > nav > nav > details > ul > li:hover > a {
	color: white;
	text-decoration: none;
	display: block;
	width: 100%;
	height: 100%;
}

body > header > a {
 	display: block;
 	text-align: center;
}


nav.top-nav {
	font-size: 16px;
	background-color: white;
 	display: flex;
 	align-items: center;
 	justify-content: space-between;
}

nav.top-nav > div {
	display: flex;
 	color: white;
}

nav.top-nav > div > form {
	padding: 1em;
}


nav.top-nav > div > form > input {
	background-color: var(--main-color);
	color: white;
 	border: 0;
 	padding: 5px 10px;
 	margin: 0;
	line-height: 24px;
}

nav.top-nav > div > form > button {
	border: 0;
 	background-color: var(--main-color);
 	padding: 5px 10px;
 	margin: 0;
	line-height: 24px;
	cursor: pointer;
}

body > main {
	margin-left: 1em;
	margin-right: 1em;
	margin-top: 1em;
	background-color: white;
	padding-left: 1em;
	padding-right: 1em;
	padding-top: 2em;
	padding-bottom: 2em;
	grid-column: 1 / -1;
}

table {
  border-collapse: collapse;
  border-color: var(--bg-color);
  border-style: solid;
  border-spacing: 0;
  background: white;
	margin-top: 2em;
	border-width: thin;
	font-size: 12px;
}

table caption {
	padding-bottom: 0.5em;
	font-size: 14px;
}

table tr:hover {
	background-color: #D0E7F9;
}

table th {
  border-color: var(--bg-color);
  padding-left: 0.5em;
  padding-right: 0.5em;
  padding-top: 0.5em;
  padding-bottom: 0.5em;
  border-style: solid;
	border-width: thin;
}

table td {
  border-color: var(--bg-color);
  padding-left: 0.5em;
  padding-right: 0.5em;
  padding-top: 0.5em;
  padding-bottom: 0.5em;
  border-style: solid;
	border-width: thin;
}

table fieldset {
	margin: 0;
	padding: 0;
	border: none;
}

table.sticky {
  position: relative;
}

table.sticky th {
  position: sticky;
  top: 0;
  background: white;
}

fieldset {
	border-right: none;
	border-left: none;
	border-width: thin;
	display: inline-grid;
	grid-template-columns: auto auto;
	column-gap: 1em;
	justify-items: left;
	grid-row-gap: 0.5em;
	margin-bottom: 1em;
	padding-bottom: 1em;
}

fieldset > fieldset {
	grid-column: 1 / -1;
}
	

legend {
	font-weight: bold;
	color: #999;
	text-transform: uppercase;
	padding-top: 1.5em;
	padding-bottom: 1.5em;
}

main input[type = text] {
	border-width: 1px;
	border-style: solid;
	border-color: #999;
	line-height: 24px;
	padding: 5px 10px;
	margin: 0 0 15px;
}

main h4 {
	font-weight: bold;
}

main ul li {
	list-style-type: none;
}

main ul li:before {
	content: '- ';
}

input[type = submit], button {
	color: white;
	padding: 5px 10px;
	border: 0;
	font-size: 14px;
	font-weight: bold;
	line-height: 24px;
	text-transform: uppercase;
	background-color: var(--main-color);
	transition: background-color 230ms ease, color 230ms ease;
}

input[type = submit]:hover, button {
	background-color: #008EC2;
	cursor: pointer;
}

input[type = submit]:disabled, button:disabled {
	color: silver;
	background-color: gray;
	cursor: not-allowed;
}

input[type = file] {
	border-style: solid;
	border-width: thin;
	border-color: silver;
}

.global {
	color: red;
	font-weight: bold;
}

.global a:link, .global a:visited {
	color: red;
	text-decoration: underline;
}
